"""
INTERNAL USE ONLY - This module is primarily intended for internal Timbal use.

This module contains functions for interacting with Timbal's Knowledge Base API.
It requires internal authentication tokens and endpoints that may not be available
in the open source distribution. Some functionality may be limited or unavailable
when used outside of the Timbal organization environment.

This is a piece of what happens when sources in a KB are automatically parsed
on the platform. The resulting extractions are uploaded to platform and linked to the original source.
"""


import mimetypes
import os
import sys
from pathlib import Path

import pandas as pd
import requests
from tqdm import tqdm


def upload_extraction(
    org_id: str,
    kb_id: str,
    source_id: str, 
    extraction_path: Path,
) -> None:
    
    content_type, _ = mimetypes.guess_type(extraction_path)
    if content_type not in ["text/markdown", "text/plain", "application/json"]:
        raise ValueError(f"Unsupported extraction content type: {content_type}")

    with open(extraction_path, "rb") as f:
        content = f.read()
    content_length = len(content)

    headers = {"Authorization": f"Bearer {os.getenv('TIMBAL_API_TOKEN')}"}
    body = {
        "content_type": content_type,
        "content_length": content_length,
    }

    res = requests.post(
        f"https://{os.getenv('TIMBAL_API_HOST')}/orgs/{org_id}/kbs/{kb_id}/sources/{source_id}/extractions",
        headers=headers,
        json=body,
    )
    res.raise_for_status()

    res_body = res.json()
    uploader = res_body.get("uploader")
    if uploader is None:
        return
    
    upload_uri = uploader.get("upload_uri")
    upload_headers = {
        "Content-Type": content_type,
        "Content-Length": str(content_length),
    }

    upload_res = requests.put(
        upload_uri,
        headers=upload_headers,
        data=content,
    )
    upload_res.raise_for_status()


def upload_dir(
    org_id: str,
    kb_id: str,
    dir_path: Path,
    dest_path: str | None = None,
) -> None:
    if not dir_path.is_dir():
        raise ValueError(f"Directory not found: {dir_path}")
    
    sources_manifest_path = dir_path / "_timbal_sources_manifest.csv"
    sources_manifest_dtypes = {
        "source_id": str,
        "is_directory": bool,
        "name": str,
        "source_path": str,
        "source_parent_path": str,
        "latest_extraction_path": str,
    }

    sources = pd.read_csv(sources_manifest_path, dtype=sources_manifest_dtypes)
    sources["relative_source_path"] = sources["source_path"] \
        .apply(lambda x: Path(x).relative_to(dir_path).with_suffix("").as_posix())

    extractions_path = dir_path / "_timbal_sources_extractions"
    extractions_ts = [int(t.stem) for t in extractions_path.iterdir() if t.is_dir()]
    latest_extractions_ts = max(extractions_ts)
    latest_extractions_path = extractions_path / str(latest_extractions_ts)

    def iterdir(path: Path):
        nonlocal sources

        for file in tqdm(path.iterdir()):
            file_posix = file.as_posix()

            if (
                file.is_dir()
                and not file.name.startswith("_") 
                and not file.name.startswith(".")
            ):
                iterdir(file)

            elif (
                file.is_file() 
                and not file.name.startswith("_") 
                and not file.name.startswith(".")
            ):
                relative_path = file.relative_to(latest_extractions_path).with_suffix("").as_posix()
                source = sources.loc[sources["relative_source_path"] == relative_path]
                if source.empty:
                    print(f"Source not found for {file_posix}")
                    continue
                source = source.iloc[0].to_dict()
                if pd.isna(source["latest_extraction_path"]) or source["latest_extraction_path"] != file_posix:
                    source_id = source["source_id"]
                    try:
                        upload_extraction(org_id, kb_id, source_id, file)
                        sources.loc[sources["source_id"] == source_id, "latest_extraction_path"] = file_posix
                    except Exception as e:
                        print(f"Error uploading extraction for {file_posix}: {e}")
                        continue

    iterdir(latest_extractions_path)

    sources.to_csv(sources_manifest_path, index=False, columns=list(sources_manifest_dtypes.keys()))


if __name__ == "__main__":
    if len(sys.argv) != 4:
        print("Usage: python -m timbal.platform.kbs.extractions <org_id> <kb_id> <dir_path>")
        sys.exit(1)

    org_id = sys.argv[1]
    kb_id = sys.argv[2]
    dir_path = Path(sys.argv[3]).expanduser().resolve()
    upload_dir(org_id, kb_id, dir_path)



"""
INTERNAL USE ONLY - This module is primarily intended for internal Timbal use.

This module contains functions for interacting with Timbal's Knowledge Base API.
It requires internal authentication tokens and endpoints that may not be available
in the open source distribution. Some functionality may be limited or unavailable
when used outside of the Timbal organization environment.

This is a preview of what will happen under the hood when you want to upload an entire
directory to a knowledge base programmatically or via the CLI.
"""


import mimetypes
import os
import sys
from pathlib import Path
from typing import Any

import pandas as pd
import requests
from tqdm import tqdm

ALLOWED_EXTENSIONS = [".pdf", ".md", ".json"]


def upload_source(
    org_id: str,
    kb_id: str,
    source_path: Path,
    parent_path: str | None = None,
    metadata: dict[str, Any] = {},
) -> str:
    content_type, _ = mimetypes.guess_type(source_path)

    with open(source_path, "rb") as f:
        content = f.read()
    content_length = len(content)

    headers = {"Authorization": f"Bearer {os.getenv('TIMBAL_API_TOKEN')}"}
    body = {
        "parent_path": parent_path,
        "is_directory": False,
        "name": source_path.name,
        "metadata": metadata,
        "content_type": content_type,
        "content_length": content_length,
    }

    res = requests.post(
        f"https://{os.getenv('TIMBAL_API_HOST')}/orgs/{org_id}/kbs/{kb_id}/sources",
        headers=headers,
        json=body,
    )
    res.raise_for_status()

    res_body = res.json()
    source_id = res_body.get("org_kb_source_id")
    uploader = res_body.get("uploader")
    if uploader is None:
        return
    
    upload_uri = uploader.get("upload_uri")
    upload_headers = {
        "Content-Type": content_type,
        "Content-Length": str(content_length),
    }

    upload_res = requests.put(
        upload_uri,
        headers=upload_headers,
        data=content,
    )
    upload_res.raise_for_status()

    return source_id


def create_dir(
    org_id: str,
    kb_id: str,
    source_name: str,
    source_parent_path: str | None = None,
    source_metadata: dict[str, Any] = {},
) -> str:
    headers = {"Authorization": f"Bearer {os.getenv('TIMBAL_API_TOKEN')}"}
    body = {
        "is_directory": True,
        "name": source_name,
        "metadata": source_metadata,
        "parent_path": source_parent_path,
        "content_type": None,
        "content_length": None,
    }

    res = requests.post(
        f"https://{os.getenv('TIMBAL_API_HOST')}/orgs/{org_id}/kbs/{kb_id}/sources",
        headers=headers,
        json=body,
    )
    res.raise_for_status()

    res_body = res.json()
    source_id = res_body.get("org_kb_source_id")
    return source_id


def upload_dir(
    org_id: str,
    kb_id: str,
    dir_path: Path,
    dest_path: str | None = None,
) -> None:
    if not dir_path.is_dir():
        raise ValueError(f"Directory not found: {dir_path}")

    
    sources_manifest_path = dir_path / "_timbal_sources_manifest.csv"
    sources_manifest_columns = [
        "source_id",                # Null if the file has not been uploaded yet.
        "is_directory",
        "name",
        "source_path",              # Local path.
        "source_parent_path",       # Built from platform parents sources ids.
        "latest_extraction_path",   # Path to the latest extraction of the source.
    ]

    if sources_manifest_path.exists():
        sources = pd.read_csv(sources_manifest_path)[sources_manifest_columns]
    else:
        sources = pd.DataFrame(columns=sources_manifest_columns)
    
    def iterdir(path, parent_path: str | None = None):
        nonlocal sources

        for file in tqdm(path.iterdir()):
            file_posix = file.as_posix()

            if (
                file.is_dir()
                and not file.name.startswith("_") 
                and not file.name.startswith(".")
            ):
                current_source = sources.loc[sources["source_path"] == file_posix]

                # If there's no existing entry in the manifest, create a new one and create the remote dir.
                if current_source.empty:
                    try:
                        source_id = create_dir(
                            org_id, 
                            kb_id, 
                            source_name=file.name, 
                            source_parent_path=parent_path, 
                            source_metadata={},
                        )
                        sources.loc[len(sources)] = {
                            "source_id": source_id,
                            "is_directory": True, 
                            "name": file.name, 
                            "source_path": file_posix,
                            "source_parent_path": parent_path,
                            "latest_extraction_path": None,
                        }
                        if pd.isna(parent_path):
                            source_parent_path = source_id
                        else:
                            source_parent_path = f"{parent_path}.{source_id}"
                        iterdir(file, parent_path=source_parent_path)
                    except Exception as e:
                        print(f"Error creating directory {file.name}: {e}")
                        continue
                else:
                    current_source = current_source.iloc[0].to_dict()
                    if pd.isna(parent_path):
                        source_parent_path = current_source["source_id"]
                    else:
                        source_parent_path = f"{parent_path}.{current_source['source_id']}"
                    iterdir(file, parent_path=source_parent_path)

            elif (
                file.is_file() 
                and file.suffix.lower() in ALLOWED_EXTENSIONS 
                and not file.name.startswith("_") 
                and not file.name.startswith(".")
            ):
                current_source = sources.loc[sources["source_path"] == file_posix]

                if current_source.empty:
                    try:
                        source_id = upload_source(
                            org_id, 
                            kb_id, 
                            source_path=file,
                            parent_path=parent_path, 
                            metadata={},
                        )
                        sources.loc[len(sources)] = {
                            "source_id": source_id,
                            "is_directory": False, 
                            "name": file.name, 
                            "source_path": file_posix,
                            "source_parent_path": parent_path,
                            "latest_extraction_path": None,
                        }
                    except Exception as e:
                        print(f"Error uploading file {file.name}: {e}")
                        continue

    iterdir(dir_path, parent_path=dest_path)

    sources.to_csv(sources_manifest_path, index=False)


if __name__ == "__main__":
    if len(sys.argv) != 4:
        print("Usage: python -m timbal.platform.kbs.sources <org_id> <kb_id> <dir_path>")
        sys.exit(1)

    org_id = sys.argv[1]
    kb_id = sys.argv[2]
    dir_path = Path(sys.argv[3]).expanduser().resolve()
    upload_dir(org_id, kb_id, dir_path)



import asyncio
import pytest
from datetime import datetime

from timbal import Agent
from timbal.errors import AgentError
from timbal.state.context import RunContext
from timbal.state.savers import InMemorySaver, JSONLSaver
from timbal.steps.perplexity import search


def get_current_time():
    return datetime.now().isoformat()


async def get_current_time_async():
    await asyncio.sleep(1)
    return datetime.now().isoformat()


def get_company_info(company_name: str):
    yield "Company name: Timbal. "
    yield "Company description: Timbal is a platform for building and running agents."


async def get_company_info_async(company_name: str):
    await asyncio.sleep(1)
    yield "Company name: Timbal. "
    await asyncio.sleep(1)
    yield "Company description: Timbal is a platform for building and running agents."


@pytest.mark.asyncio
async def test_init():
    agent = Agent()
    print(agent)


@pytest.mark.asyncio
async def test_init_with_tools():
    agent = Agent(tools=[get_current_time, search])
    print("Tools: ", agent.tools)
    print("Tools lookup: ", agent.tools_lookup)
    for tool in agent.tools:
        print(tool.to_openai_tool())
        print(tool.to_anthropic_tool())


@pytest.mark.asyncio
async def test_params_model():
    agent = Agent()
    print(agent.params_model())


@pytest.mark.asyncio
async def test_params_model_schema():
    agent = Agent()
    print(agent.params_model_schema())


@pytest.mark.asyncio
async def test_return_model():
    agent = Agent()
    print(agent.return_model())


@pytest.mark.asyncio
async def test_return_model_schema():
    agent = Agent()
    print(agent.return_model_schema())


@pytest.mark.asyncio
async def test_run():
    agent = Agent()
    async for event in agent.run(prompt="Hello, world!"):
        print(event)


@pytest.mark.asyncio
async def test_run_with_sync_tool():
    agent = Agent(tools=[get_current_time])
    await agent.complete(prompt="What time is it?")


@pytest.mark.asyncio
async def test_run_with_async_tool():
    agent = Agent(tools=[get_current_time_async])
    async for event in agent.run(prompt="What time is it?"):
        print(event)


@pytest.mark.asyncio
async def test_run_with_sync_gen_tool():
    agent = Agent(tools=[get_company_info])
    async for event in agent.run(prompt="Give me the info of the company Timbal."):
        print(event)


@pytest.mark.asyncio
async def test_run_with_async_gen_tool():
    agent = Agent(tools=[get_company_info_async])
    async for event in agent.run(prompt="Give me the info of the company Timbal."):
        print(event)


@pytest.mark.asyncio
async def test_run_with_multiple_tools():
    agent = Agent(tools=[get_current_time, get_company_info_async])
    async for event in agent.run(prompt="Give me the time and the info of the company Timbal."):
        print(event)


def get_customer(customer_id):
    if isinstance(customer_id, int):
        raise ValueError("Customer ID must be a string.")
    return {"id": customer_id, "name": "John Doe", "email": "john.doe@example.com"}


@pytest.mark.asyncio
async def test_run_incomplete_type_annotations():
    agent = Agent(tools=[get_customer])
    await agent.complete(prompt="Give me the info of the customer 123")


@pytest.mark.asyncio
async def test_run_with_max_iter():
    agent = Agent(tools=[get_customer], max_iter=1)
    async for event in agent.run(prompt="Give me the info of the customer 123"):
        print(event)


async def get_weather(city: str):
    yield "The weather in " 
    await asyncio.sleep(3)
    yield city
    await asyncio.sleep(3)
    yield " is sunny."


@pytest.mark.asyncio
async def test_run_parallel_tool_calls():
    agent = Agent(tools=[get_weather])
    async for event in agent.run(prompt="What is the weather in Tokyo and in London?"):
        print(event)


@pytest.mark.asyncio
async def test_run_with_memory():
    agent = Agent(
        tools=[get_weather],
        state_saver=InMemorySaver(),
    )

    async for event in agent.run(prompt="What is the weather in Tokyo and in London?"):
        parent_run_id = event.run_id
        print()
        print("Event: ", event)

    async for event in agent.run(
        context=RunContext(parent_id=parent_run_id),
        prompt="What was I talking about?"
    ):
        print()
        print("Event: ", event)


@pytest.mark.asyncio
async def test_run_with_parallel_internet_search():
    agent = Agent(tools=[search])

    async for event in agent.run(prompt="What is the weather in Tokyo and in London?"):
        print()
        print("Event: ", event)


@pytest.mark.asyncio
async def test_run_llm_error():
    agent = Agent(model="this-model-does-not-exist")
    with pytest.raises(AgentError):
        async for event in agent.run(prompt="What is the weather in Tokyo and in London?"):
            print()
            print("Event: ", event)


# TODO Test initial kwargs validation error.
# TODO Test data is being saved properly after any type of error.


@pytest.mark.asyncio
async def test_other_models():
    state_saver = JSONLSaver("test_jsonl.jsonl")
    agent = Agent(
        model="gpt-4o-mini",
        # model="gemini-2.0-flash-lite",
        # model="o3-mini",
        # model="Qwen/Qwen2.5-7B-Instruct-Turbo",
        # model="meta-llama/Llama-4-Scout-17B-16E-Instruct",
        # model="claude-3-5-sonnet-20241022",
        # model="o1",
        tools=[get_current_time],
        state_saver=state_saver,
        max_tokens=2048,
    )
    res = await agent.complete(prompt="What is the time?")
    print(res)


@pytest.mark.asyncio
async def test_parallel_search():
    # TODO Since gemini does not return an id the paths look like this:  path='agent.search-'
    agent = Agent(
        model="gemini-2.0-flash-lite",
        tools=[{
            "runnable": search,
            "description": "Search the weather on the internet.",
            "params_mode": "required",
        }],
    )
    res = await agent.complete(prompt="What is the weather in Tokyo and in London?")
    print(res)



import os
from pathlib import Path
from typing import Any

import pandas as pd
import requests

# class ChunkInput(BaseModel):
#     """"""
#     id: str
#     source_id: str
#     content: str
#     metadata: dict[str, Any]


def add_chunks(
    org_id: str,
    kb_id: str,
    chunks: list[Any],
) -> None:
    headers = {"Authorization": f"Bearer {os.getenv('TIMBAL_API_TOKEN')}"}
    body = {"chunks": chunks}

    res = requests.post(
        f"https://{os.getenv('TIMBAL_API_HOST')}/orgs/{org_id}/kbs/{kb_id}/sources/chunks",
        headers=headers,
        json=body,
    )
    res.raise_for_status()


if __name__ == "__main__":
    chunks_path = Path("/Users/dberges/Desktop/timbal-ai/kbs/22/_timbal_sources_chunks/1743692857559.csv")
    df = pd.read_csv(chunks_path)[["source_id", "content"]]
    df = df[df["source_id"].notna()].reset_index(drop=True)
    df["source_id"] = df["source_id"].astype(int).astype(str)

    chunks = [
        {
            "source_id": row["source_id"],
            "content": row["content"],
            "metadata": {},
        }
        for _, row in df.iterrows()
    ]

    add_chunks(org_id="9", kb_id="22", chunks=chunks)



from pydantic import BaseModel, ConfigDict


class BaseEvent(BaseModel):
    """Base class for all timbal events yielded during flow execution."""
    # Allow storing extra fields in the model.
    model_config = ConfigDict(extra="allow")

    type: str
    """The type of the event. This will be very useful for serializing and deserializing events."""
    run_id: str
    """The id of the run this event was emitted from."""
    path: str
    """The path of the element that yielded this event."""



import pytest
import os
from pathlib import Path

from timbal import Agent
from timbal.state.savers.jsonl import JSONLSaver
from timbal.types import File
from timbal.steps.elevenlabs import tts # , stt

from timbal.state.context import RunContext


# No need to run these explicitly, since they'll always be run internally by all the agents tests.
# @pytest.mark.asyncio
# async def test_stt():
#     audio_file = File.validate("https://cdn.openai.com/API/docs/audio/alloy.wav")

#     res = await stt(audio_file=audio_file)
#     print(res)


# @pytest.mark.asyncio
# async def test_tts():
#     text = "Hello, how are you?"
#     voice_id = "56AoDkrOh6qfVPDXZ7Pt"

#     res = await tts(text=text, voice_id=voice_id)
#     run_context = RunContext.model_validate({
#         "id": "test",
#         "timbal_platform_config": {
#             "host": "dev.timbal.ai",
#             "auth_config": {
#                 "type": "bearer",
#                 "token": os.getenv("TIMBAL_API_TOKEN"),
#             },
#             "app_config": {
#                 "org_id": "1",
#                 "app_id": "18",
#                 "version_id": "85",
#             },
#         },
#     })
#     res_ser = File.serialize(res, run_context) 
#     print(res_ser)


@pytest.fixture(params=[
    Path(__file__).parent / "fixtures" / "alloy.wav",
    "https://cdn.openai.com/API/docs/audio/alloy.wav",
])
def audio(request):
    return request.param


# TODO Test all of this with multiple models.
@pytest.mark.asyncio
async def test_agent_stt(audio):
    audio_file = File.validate(audio)

    agent = Agent(model="gpt-4.1-nano")

    res = await agent.complete(prompt=audio_file)
    print(res)


# TODO Change this.
@pytest.mark.asyncio
async def test_agent_stt_tts():
    # audio_file = File.validate("https://cdn.openai.com/API/docs/audio/alloy.wav")

    agent = Agent(
        model="gpt-4o-mini",
        tools=[{
            "runnable": tts,
            "force_exit": True
        }],
        system_prompt=(
            "You are a helpful assistant and you must always respond in audio format. "
            "Always use '56AoDkrOh6qfVPDXZ7Pt' as the voice_id for the TTS model."
        ),
        state_saver=JSONLSaver("test_jsonl.jsonl")
    )

    run_context = RunContext.model_validate({
        "id": "test",
        "timbal_platform_config": {
            "host": "dev.timbal.ai",
            "auth_config": {
                "type": "bearer",
                "token": os.getenv("TIMBAL_API_TOKEN"),
            },
            "app_config": {
                "org_id": "1",
                "app_id": "18",
                "version_id": "85",
            },
        },
    })

    res = await agent.complete(prompt="Hello how are you?", context=run_context)



import os

from timbal import Agent
from timbal.core.shared import RemoteConfig

from timbal.steps.timbal.tables import search_table

# For testing purposes
# os.environ["TIMBAL_API_HOST"] = "api.timbal.ai"


async def main():
    # remote_config = RemoteConfig(
    #     org_id="1",
    #     app_id="49",
    # )
    # agent = Agent(remote_config=remote_config)

    # # # You can run this in as a stream
    # # async for event in agent.run(prompt="Hello"):
    # #     print(event)

    # # Or you can run this as a whole and get the final output event
    # res = await agent.complete(prompt="Hello")
    # print(res)

    res = await search_table(
        org_id="10",
        kb_id="48",
        table_name="Documents",
        query="digam el tratamiento preventivo del BIOESTATICO COMBUSTIBLE",
        embedding_names=["esther-demo"],
        limit=1,
    )
    print(res)


if __name__ == "__main__":
    import asyncio
    asyncio.run(main())